/* ===========================================================================
 * Modul: zg2s.c
 *
 * Version 1.1
 *
 * Copyright (C) 2005/2006 Landesumweltamt  NRW, Germany
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 *----------------------------------------------------------------------------
 *
 * Aus den Gelaendehoehen, die in einer dmna-Datei stehen,
 * werden die Steigungen ermittelt. Diese werden in Form
 * einer dmna-Datei ausgegeben.
 * ==========================================================================*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <ctype.h>
#include "zg2s.h"

void headerInit(void);
int argumenteEinlesen(int argc, char *argv[]);
int datenEinlesen(void);
void steigungBerechnen(void);
int dmnaDateiErstellen(void);
void msg(int rc);
void feldFree(void);

static const char 
   info[] = "\n"
      "Copyright (C) 2005/2006  Landesumweltamt NRW, Germany\n\n"
      "Version 1.1\n",
   aufruf[] = "\n"
      "Programm zur Bestimmung der Steigungen von gegebenen Gelaendehoehen\n\n"
      "Aufruf:     zg2s <dmna-Datei>\n\n"
      "dmna-Datei: dmna-Datei mit den Gelaendehoehen (z. B. zg00.dmna)\n\n"
      "Die Datei mit dem Ergebnis bekommt den Namen der dmna-Datei,\n"
      "der aber um \"_s\" ergaenzt wird (z. B. zg00_s.dmna).\n";

static char src[256], dst[256];
static RH **feld;
static struct header h;

static int xVertikal = -1,
           xDim, yDim,
           xSign = 1, ySign = 1;

int main(int argc, char *argv[])
{
   int rc = 0;

   msg(MSG_INFO);
   if(!(rc = argumenteEinlesen(argc, argv)))
   {
      if(!(rc = datenEinlesen()))
      {
         steigungBerechnen();
         rc = dmnaDateiErstellen();
      }
      feldFree();
   }
   msg(rc);
   printf("\n*** Programmende ***\n");

   return rc;
}

void headerInit(void)
{
   h.dims = 2;
   h.size = 4;
   strcpy(h.locl, "\"C\"");
   strcpy(h.form, "\"%5.2f\"");
   strcpy(h.artp, "\"T\"");
   strcpy(h.vldf, "\"P\"");
   strcpy(h.sequ, "\"j-,i+\"");
   strcpy(h.mode, "\"text\"");
   h.lowb[0] = h.lowb[1] = -1;
   h.hghb[0] = h.hghb[1] = -1;
   h.xmin = 0.0;
   h.ymin = 0.0;
   h.gakrx = 0.0;
   h.gakry = 0.0;
   h.delta = 0.0;
}

int argumenteEinlesen(int argc, char *argv[])
{
   int c, rc = 0;

   if(argc != 2)
   {
      rc = MSG_AUFRUF;
   }
   else
   {
      c = (*++argv)[0];
      if(c == '-' || c == '/')
      {
         rc = OPT_KEINE;
      }
      else
         strcpy(src, *argv);
   }
   return rc;
}

int datenEinlesen(void)
{
   int  i, j, rc = 0;
   double xMin, yMin;
   FILE *fs;
   char *p, s[512];
   
   if((fs = fopen(src, "r")) == NULL)
      return MSG_DMNA_FILE_IFS;
   printf("\ndmna-Datei  %s  wird gelesen.\n", src);

   headerInit();
	while(fgets(s, sizeof(s), fs))
	{
      if((p = strchr(s, '*')) != NULL)
      {
         for(i = 0, p = s; *p; ++p)
            if(!isspace(*p))
               ++i;
         if(i == 1)
            break;
      }
      if((p = strchr(s, '\n')) != NULL)
         *p = '\0';
      p = strlwr(s);
      while(isspace(*p))
         ++p;
      if(*p == '-')
         continue;
      if((p = strchr(s, '\'')) != NULL)
         *p = '\0';
      if((p = strstr(s, "sequ")) != NULL)
      {
         if((p = strpbrk(p+4, " \t\"")) != NULL)
			{
				while(isspace(*p))
				   ++p;
				if(strlen(p) >= 5)
				{
					strcpy(h.sequ, p);
					if(p[0] == '"')
						++p;
					if(p[0] == 'j' && p[3] == 'i')
					{
						xVertikal = 0;
						ySign = p[1] == '-' ? -1 : 1;
						xSign = p[4] == '-' ? -1 : 1;
					}
					if(p[0] == 'i' && p[3] == 'j')
					{
						xVertikal = 1;
						xSign = p[1] == '-' ? -1 : 1;
						ySign = p[4] == '-' ? -1 : 1;
					}
				}
				else
					return MSG_SEQU;
			}
      }
      if((p = strstr(s, "xmin")) != NULL)
      {
         if((p = strpbrk(p+4, " \t")) != NULL)
            h.xmin = atof(p);
      }
      if((p = strstr(s, "ymin")) != NULL)
      {
         if((p = strpbrk(p+4, " \t")) != NULL)
            h.ymin = atof(p);
      }
      if((p = strstr(s, "delta")) != NULL)
      {
         if((p = strpbrk(p+5, " \t")) != NULL)
            h.delta = atof(p);
      }
      if((p = strstr(s, "gakrx")) != NULL)
      {
         if((p = strpbrk(p+5, " \t")) != NULL)
            h.gakrx = atof(p);
      }
      if((p = strstr(s, "gakry")) != NULL)
      {
         if((p = strpbrk(p+5, " \t")) != NULL)
            h.gakry = atof(p);
      }
      if((p = strstr(s, "lowb")) != NULL)
      {
         sscanf(p, "%*s%d%d", &h.lowb[0], &h.lowb[1]);
      }
      if((p = strstr(s, "hghb")) != NULL)
      {
         sscanf(s, "%*s%d%d", &h.hghb[0], &h.hghb[1]);
      }
   }
   xMin = h.gakrx + h.xmin;
   yMin = h.gakry + h.ymin;
   if(!strlen(h.sequ) || xVertikal == -1)
      return MSG_SEQU;
   if(h.delta <= 0.0)
      return MSG_DELTA;

   xDim = h.hghb[0] - h.lowb[0] + 1;
   yDim = h.hghb[1] - h.lowb[1] + 1;
   if(xDim > 2 && yDim > 2)
   {
      if((feld = (RH **)calloc(xDim, sizeof(RH *))) != NULL)
      {
         for(i = 0; i < xDim && !rc; ++i)
            rc = (feld[i] = (RH *)calloc(yDim, sizeof(RH))) == NULL ? MSG_MEM : 0;
      }
      else
         rc = MSG_MEM;
      if(rc != MSG_MEM)
		{
         if(xVertikal)
         {
            for(i = xSign > 0 ? 0 : xDim-1; i < xDim && i >= 0; i += xSign)
            {
               for(j = ySign > 0 ? 0 : yDim-1; j < yDim && j >= 0; j += ySign)
               {
                  fscanf(fs, "%s", s);
                  feld[i][j].hoehe = atof(s);
               }
            }
         }
         else
         {
            for(j = ySign > 0 ? 0 : yDim-1; j < yDim && j >= 0; j += ySign)
            {
               for(i = xSign > 0 ? 0 : xDim-1; i < xDim && i >= 0; i += xSign)
               {
                  fscanf(fs, "%s", s);
                  feld[i][j].hoehe = atof(s);
               }
            }
         }
      }
   }
   else
      rc = MSG_LHB;
   fclose(fs);
   return rc;
}

void steigungBerechnen(void)
{
   int i, j;
   double abstand, st;

   abstand = 2.0 * h.delta;
   for(j = 1; j < yDim - 1; ++j)
   {
      for(i = 1; i < xDim - 1; ++i)
      {
         st = fabs(feld[i-1][j].hoehe - feld[i+1][j].hoehe) / abstand;
         if(feld[i][j].steigung < st)
            feld[i][j].steigung = st;
         st = fabs(feld[i][j-1].hoehe - feld[i][j+1].hoehe) / abstand;
         if(feld[i][j].steigung < st)
            feld[i][j].steigung = st;
      }
   }

   for(j = 1; j < yDim - 1; ++j)
      for(i = 1; i < xDim - 1; ++i)
         feld[i][j].klasse =
            (feld[i][j].steigung <= 0.05) ? KL1 : ((feld[i][j].steigung <= 0.2) ? KL2 : KL3);
}

int dmnaDateiErstellen(void)
{
   int i, j;
   FILE *fd;
   char *p;

   strcpy(dst, src);
   if((p = strrchr(dst, '.')) != NULL)
      *p = '\0';
   strcat(dst, "_s.dmna");

   if((fd = fopen(dst, "w")) == NULL)
      return MSG_DMNA_FILE_OFS;
   printf("dmna-Datei  %s  wird erstellt.\n", dst);
   /* Header */
   fprintf(fd, "locl\t%s\nxmin\t%.1lf\nymin\t%.1lf\ndelta\t%.1lf\ngakrx\t%.1lf\ngakry\t%.1lf\n"
               "form\t%s\nartp\t%s\nvldf\t%s\nsequ\t%s\nmode\t%s\ndims\t%d\n"
               "size\t%d\nlowb\t%d\t%d\nhghb\t%d\t%d\n*\n",
               h.locl, h.xmin+h.delta, h.ymin+h.delta,  h.delta, h.gakrx, h.gakry,
               h.form, h.artp, h.vldf, h.sequ, h.mode, h.dims, h.size,
               h.lowb[0]+1, h.lowb[1]+1, h.hghb[0]-1, h.hghb[1]-1);

   if(xVertikal)
   {
      for(i = xSign > 0 ? 1 : xDim-2; i < xDim-1 && i >= 1; i += xSign)
      {
         for(j = ySign > 0 ? 1 : yDim-2; j < yDim-1 && j >= 1; j += ySign)
         {
            fprintf(fd, "\t%5.2lf", feld[i][j].steigung);
         }
         fprintf(fd, "\n");
      }
   }
   else
   {
      for(j = ySign > 0 ? 1 : yDim-2; j < yDim-1 && j >= 1; j += ySign)
      {
         for(i = xSign > 0 ? 1 : xDim-2; i < xDim-1 && i >= 1; i += xSign)
         {
            fprintf(fd, "\t%5.2lf", feld[i][j].steigung);
         }
         fprintf(fd, "\n");
      }
   }
   fprintf(fd, "\n***\n");
   fclose(fd);
   return 0;
}

void msg(int rc)
{
   switch(rc)
   {
      case MSG_INFO:
         printf("%s", info);
         break;
      case MSG_AUFRUF:
         printf("%s", aufruf);
         break;
      case MSG_DMNA_FILE_IFS:
         printf("\nDmna-Datei kann nicht gelesen werden.\n");
         break;
      case MSG_DMNA_FILE_OFS:
         printf("\nDmna-Datei kann nicht erstellt werden.\n");
         break;
      case MSG_DELTA:
         printf("\nUngueltiger Delta-Wert\n");
         break;
      case MSG_SEQU:
         printf("\nUngueltige Angabe fuer sequ\n");
         break;
      case MSG_LHB:
         printf("\nUngueltige Werte fuer lowb bzw. hghb\n");
         break;
      case MSG_MEM:
         printf("\nSpeicherallokierung ist fehlgeschlagen.\n");
         break;
      case OPT_KEINE:
         printf("\nOptionen werden nicht erwartet.\n");
         break;
   }
}

void feldFree(void)
{
   int i;
   
   for(i = 0; i < xDim; ++i)
      if(feld[i])
         free(feld[i]);
   if(feld)
      free(feld);
}
